import {Meta} from '@storybook/react/types-6-0';
import faker from 'faker';
import moment from 'moment';
import * as React from 'react';

import {RunFilterToken} from '../runs/RunsFilterInput';
import {StorybookProvider} from '../testing/StorybookProvider';
import {RunStatus, StepEventStatus} from '../types/globalTypes';

import {PartitionRunMatrix} from './PartitionRunMatrix';
import {
  PartitionSetLoaderRunFragment,
  PartitionSetLoaderRunFragment_tags,
} from './types/PartitionSetLoaderRunFragment';

// eslint-disable-next-line import/no-default-export
export default {
  title: 'PartitionRunMatrix',
  component: PartitionRunMatrix,
} as Meta;

interface ISolidHandle {
  __typename: 'SolidHandle';
  handleID: string;
  solid: {
    name: string;
    definition: {__typename: 'SolidDefinition'; name: string};
    inputs: {dependsOn: {solid: {name: string}}[]}[];
    outputs: {dependedBy: {solid: {name: string}}[]}[];
  };
}

// On the real PartitionView, the tags in the filter bar are passed as a GraphQL query param,
// to make them work just filter our sample data.
function simulateTagFilteringInQuery(
  runs: PartitionSetLoaderRunFragment[],
  runFilters: RunFilterToken[],
) {
  return runs.filter((r) =>
    runFilters.every((t) => r.tags.some((rt) => `${rt.key}=${rt.value}` === t.value)),
  );
}

function buildOpHandles(opNames: string[], deps: string[]) {
  const result: ISolidHandle[] = [];
  for (const name of opNames) {
    result.push({
      __typename: 'SolidHandle',
      handleID: name,
      solid: {
        name: name,
        definition: {
          __typename: 'SolidDefinition',
          name: name,
        },
        inputs: [],
        outputs: [],
      },
    });
  }
  for (const dep of deps) {
    const [_, from, to] = /([^\s]*)\s*=>\s*([^\s]*)/.exec(dep)!;
    result
      .find((s) => s.handleID === from)!
      .solid.outputs.push({
        dependedBy: [{solid: {name: to}}],
      });
    result
      .find((s) => s.handleID === to)!
      .solid.inputs.push({
        dependsOn: [{solid: {name: from}}],
      });
  }
  return result;
}

function buildRun(
  isoDateString: string,
  statuses: {[stepKey: string]: StepEventStatus},
  tags: PartitionSetLoaderRunFragment_tags[] = [],
) {
  const id = faker.datatype.uuid().slice(0, 8);
  const startTime = new Date(isoDateString).getTime() / 1000;
  const result: PartitionSetLoaderRunFragment = {
    __typename: 'Run',
    id: id,
    runId: id,
    status: RunStatus.SUCCESS,
    stats: {
      id: id,
      startTime,
      endTime: startTime + 1,
      materializations: 0,
      __typename: 'RunStatsSnapshot',
    },
    stepStats: Object.entries(statuses).map(([key, status]) => ({
      __typename: 'RunStepStats',
      stepKey: key,
      status: status,
      startTime: startTime,
      endTime: startTime + 1,
      materializations: [],
      expectationResults: [],
    })),
    tags: tags,
  };

  return result;
}

const PipelineMocks = {
  PipelineSnapshotOrError: () => ({
    __typename: 'PipelineSnapshot',
  }),
  PipelineSnapshot: () => ({
    name: () => 'TestPipeline',
    solidHandles: () =>
      buildOpHandles(
        ['first', 'second', 'third', 'fourth'],
        ['first => second', 'first => third', 'second => fourth', 'third => fourth'],
      ),
  }),
};

export const BasicTestCases = () => {
  const [runFilters, setRunFilters] = React.useState<RunFilterToken[]>([]);
  const [stepQuery, setStepQuery] = React.useState('');
  const partitions: {name: string; runs: PartitionSetLoaderRunFragment[]}[] = [];

  // Good Run
  partitions.push({
    name: '2021-01-01',
    runs: [
      buildRun('2021-01-01T00:00:00', {
        first: StepEventStatus.SUCCESS,
        second: StepEventStatus.SUCCESS,
        third: StepEventStatus.SUCCESS,
        fourth: StepEventStatus.SUCCESS,
      }),
    ],
  });
  // Partial Failure Run
  partitions.push({
    name: '2021-01-02',
    runs: [
      buildRun('2021-01-02T00:00:00', {
        first: StepEventStatus.SUCCESS,
        second: StepEventStatus.FAILURE,
        third: StepEventStatus.SUCCESS,
        fourth: StepEventStatus.SKIPPED,
      }),
    ],
  });
  // Some Steps Skipped Run
  partitions.push({
    name: '2021-01-03',
    runs: [
      buildRun('2021-01-03T00:00:00', {
        first: StepEventStatus.SUCCESS,
        second: StepEventStatus.SUCCESS,
        third: StepEventStatus.SKIPPED,
        fourth: StepEventStatus.SKIPPED,
      }),
    ],
  });
  // Still Running Run
  partitions.push({
    name: '2021-01-04',
    runs: [
      buildRun('2021-01-04T00:00:00', {
        first: StepEventStatus.SUCCESS,
      }),
    ],
  });

  // Day with No Runs
  partitions.push({
    name: '2021-01-05',
    runs: [],
  });

  // Partial Failure Run followed by "Restart from Failure" Run
  partitions.push({
    name: '2021-01-06',
    runs: [
      buildRun('2021-01-06T00:00:00', {
        first: StepEventStatus.SUCCESS,
        second: StepEventStatus.FAILURE,
        third: StepEventStatus.SKIPPED,
      }),
      buildRun(
        '2021-01-06T12:00:00',
        {
          second: StepEventStatus.SUCCESS,
          third: StepEventStatus.SUCCESS,
          fourth: StepEventStatus.SUCCESS,
        },
        [{__typename: 'PipelineTag', key: 'dagster/backfill', value: '1234'}],
      ),
    ],
  });

  // Successful run followed by a run with failure and skips
  partitions.push({
    name: '2021-01-07',
    runs: [
      buildRun('2021-01-07T00:00:00', {
        first: StepEventStatus.SUCCESS,
        second: StepEventStatus.SUCCESS,
        third: StepEventStatus.SUCCESS,
        fourth: StepEventStatus.SUCCESS,
      }),
      buildRun(
        '2021-01-07T12:00:00',
        {
          first: StepEventStatus.SUCCESS,
          second: StepEventStatus.FAILURE,
          third: StepEventStatus.SKIPPED,
        },
        [
          {__typename: 'PipelineTag', key: 'dagster/backfill', value: '1234'},
          {__typename: 'PipelineTag', key: 'randomtag', value: 'helloworld'},
        ],
      ),
    ],
  });

  return (
    <StorybookProvider apolloProps={{mocks: PipelineMocks}}>
      <PartitionRunMatrix
        pipelineName="TestPipeline"
        repoAddress={{name: 'Test', location: 'TestLocation'}}
        runFilters={runFilters}
        setRunFilters={setRunFilters}
        stepQuery={stepQuery}
        setStepQuery={setStepQuery}
        partitions={partitions.map((p) => ({
          ...p,
          runsLoaded: true,
          runs: simulateTagFilteringInQuery(p.runs, runFilters),
        }))}
      />
    </StorybookProvider>
  );
};

export const LargeDataset = () => {
  const [runFilters, setRunFilters] = React.useState<RunFilterToken[]>([]);
  const [stepQuery, setStepQuery] = React.useState('');
  const partitions = React.useMemo(() => {
    const results: {name: string; runs: PartitionSetLoaderRunFragment[]}[] = [];
    for (let ii = 0; ii < 300; ii++) {
      const date = moment().startOf('year').add(ii, 'days');
      const runCount = faker.random.number(10);

      const runs: PartitionSetLoaderRunFragment[] = [];
      for (let x = 0; x < runCount; x++) {
        runs.push(
          buildRun(date.clone().add(x, 'minutes').toISOString(), {
            first: faker.random.objectElement(StepEventStatus) as any,
            second: faker.random.objectElement(StepEventStatus) as any,
            third: faker.random.objectElement(StepEventStatus) as any,
            fourth: faker.random.objectElement(StepEventStatus) as any,
          }),
        );
      }
      results.push({
        name: date.format('YYYY-MM-DD'),
        runs: runs,
      });
    }
    return results;
  }, []);

  return (
    <StorybookProvider apolloProps={{mocks: PipelineMocks}}>
      <PartitionRunMatrix
        pipelineName="TestPipeline"
        repoAddress={{name: 'Test', location: 'TestLocation'}}
        runFilters={runFilters}
        setRunFilters={setRunFilters}
        stepQuery={stepQuery}
        setStepQuery={setStepQuery}
        partitions={partitions.map((p) => ({
          ...p,
          runsLoaded: true,
          runs: simulateTagFilteringInQuery(p.runs, runFilters),
        }))}
      />
    </StorybookProvider>
  );
};
